#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains default implementation for tpRigToolkit configurations
"""

from __future__ import print_function, division, absolute_import

__author__ = "Tomas Poveda"
__license__ = "MIT"
__maintainer__ = "Tomas Poveda"
__email__ = "tpovedatd@gmail.com"

import os
import logging

import metayaml

import tpDcc as tp

LOGGER = logging.getLogger()


class YAMLConfigurationAttribute(dict, object):
    """
    Class that allows access nested dictionaries using Python attribute access
    https://stackoverflow.com/questions/38034377/object-like-attribute-access-for-nested-dictionary
    """

    def __init__(self, *args, **kwargs):
        super(YAMLConfigurationAttribute, self).__init__(*args, **kwargs)
        self.__dict__ = self

    @staticmethod
    def from_nested_dict(data):
        """
        Construst a nested YAMLConfigurationAttribute from nested dictionaries
        :param data: dict
        :return: YAMLConfigurationAttribute
        """

        if not isinstance(data, dict):
            return data
        else:
            return YAMLConfigurationAttribute(
                {key: YAMLConfigurationAttribute.from_nested_dict(data[key]) for key in data})


class YAMLConfigurationParser(object):
    def __init__(self, config_data):
        super(YAMLConfigurationParser, self).__init__()

        self._config_data = config_data
        self._parsed_data = dict()

    def parse(self):
        self._parsed_data = self._config_data
        return YAMLConfigurationAttribute.from_nested_dict(self._parsed_data)


class YAMLConfiguration(object):
    def __init__(self, config_name, environment, config_dict=None,
                 parser_class=YAMLConfigurationParser):
        super(YAMLConfiguration, self).__init__()

        self._config_name = config_dict
        self._environment = environment
        self._parser_class = parser_class
        self._parsed_data = self.load(config_name=config_name, config_dict=config_dict)

    @property
    def data(self):
        return self._parsed_data

    def get_path(self):
        if not self._parsed_data:
            return None

        return self._parsed_data.get('config', {}).get('path', None)

    def get(self, attr_section, attr_name=None, default=None):
        """
        Returns an attribute of the configuration
        :param attr_name: str
        :param attr_section: str
        :param default: object
        :return:
        """

        if not self._parsed_data:
            LOGGER.warning('Configuration "{}" is empty for "{}"'.format(
                self._config_name, self._environment))
            return default

        if attr_section and attr_name:
            orig_section = attr_section
            attr_section = self._parsed_data.get(attr_section, dict())
            if not attr_section:
                LOGGER.warning('Configuration "{}" has no attribute "{}" in section "{}" for "{}"'.format(
                    self._config_name, attr_name, orig_section, self._environment))
                return default
            attr_value = attr_section.get(attr_name, None)
            if attr_value is None:
                LOGGER.warning('Configuration "{}" has no attribute "{}" in section "{}" for "{}"'.format(
                    self._config_name, attr_name, attr_section, self._environment))
                return default
            return attr_value
        else:
            attr_to_use = attr_section
            if attr_name and not default:
                default = attr_name
            if not attr_section:
                attr_to_use = attr_name
            attr_value = self._parsed_data.get(attr_to_use, None)
            if attr_value is None:
                LOGGER.warning('Configuration "{}" has no attribute "{}" for "{}"'.format(
                    self._config_name, attr_to_use, self._environment))
                return default
            return attr_value

    def load(self, config_name, config_dict=None):
        """
        Function that reads configuration file and initializes project variables properly
        This function can be extended in new projects
        """

        if not config_dict:
            config_dict = {}

        config_data = self._get_config_data(config_name, config_dict=config_dict)
        if not config_data:
            return False

        parsed_data = self._parser_class(config_data).parse()

        return parsed_data

    def _get_tprigtoolkit_config_paths(self, module_config_name):
        """
        Returns a list of paths where configuration file can be located in tpRigToolkit-config
        :return: list(str)
        """

        configs_path = os.environ.get('TPRIGTOOLKIT_CONFIGS_PATH', None)
        if not configs_path or not os.path.isdir(configs_path):
            from tpRigToolkit import config
            config_dir = os.path.dirname(config.__file__)
        else:
            config_dir = configs_path

        config_env_dir = os.path.join(config_dir, self._environment.lower())
        if not os.path.isdir(config_env_dir):
            LOGGER.warning('Configuration Folder for Environment "{}" does not exists: "{}"'.format(
                self._environment, config_env_dir))

        config_path = os.path.join(config_env_dir, module_config_name)
        dcc_config_path = os.path.join(config_env_dir, tp.Dcc.get_name(), module_config_name)
        dcc_version_config_path = os.path.join(
            config_env_dir, tp.Dcc.get_name(), tp.Dcc.get_version_name(), module_config_name)

        return [config_path, dcc_config_path, dcc_version_config_path]

    def _get_config_paths(self, module_config_name, skip_non_existent=True):
        """
        Returns a list of valid paths where configuration files can be located
        :return: list(str)
        """

        found_paths = list()
        paths_to_found = self._get_tprigtoolkit_config_paths(module_config_name=module_config_name)

        for p in paths_to_found:
            if not p:
                continue
            if skip_non_existent and not os.path.isfile(p):
                continue
            found_paths.append(p)

        return found_paths

    def _get_config_data(self, config_name, config_dict):
        """
        Returns the config data
        :return: dict
        """

        if not config_name:
            tp.Dcc.error(
                'Configuration File for tpRigToolkit not found! {}'.format(self, config_name))
            return

        module_config_name = config_name + '.yml'

        # We use tpRigToolkit configuration as base configuration file
        all_config_paths = self._get_config_paths(module_config_name=module_config_name, skip_non_existent=False)
        valid_config_paths = self._get_config_paths(module_config_name=module_config_name)
        if not valid_config_paths:
            raise RuntimeError(
                'Impossible to load configuration "{}"  for tpRigToolkit because it does not exists in any of '
                'the configuration folders: {}'.format(config_name, ''.join(all_config_paths)))

        config_path = valid_config_paths[-1]
        config_data = metayaml.read(valid_config_paths, config_dict)
        if not config_data:
            raise RuntimeError(
                'Configuration File is empty! {}'.format(self, config_path))

        # We store path where configuration file is located in disk
        if 'config' in config_data and 'path' in config_data['config']:
            raise RuntimeError(
                'Configuration File for cannot contains config section with path attribute! {}'.format(
                    self, config_path))
        if 'config' in config_data:
            config_data['config']['path'] = config_path
        else:
            config_data['config'] = {'path': config_path}

        return config_data


def get_config(config_name, config_dict=None, parser_class=YAMLConfigurationParser):
    """
    Returns a configuration with the given arguments
    :param config_name: str
    :param config_dict: dict
    :param parser_class: YAMLConfigurationParser
    :param environment: str (optional)
    :return: YAMLConfiguration
    """

    new_cfg = YAMLConfiguration(
        config_name=config_name,
        config_dict=config_dict,
        parser_class=parser_class,
        environment=os.environ.get('tpRigToolkit_env', 'DEVELOPMENT')
    )

    return new_cfg



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains data manager widgets
"""

from __future__ import print_function, division, absolute_import

import os
import pkgutil
import inspect
import traceback

from tpDcc.core import data as core_data, scripts
from tpDcc.libs.python import decorators

import tpRigToolkit


class ScriptsManager(object):

    def __init__(self):
        super(ScriptsManager, self).__init__()

        self.directories = list()
        self.ask_name_on_creation = True
        self.loaded_data_classes = list()

        self.standard_data_classes = [
            scripts.ScriptManifestData,
            scripts.ScriptPythonData
        ]

    def get_all_data_classes(self, _reload=False):
        """
        Returns all data widgets loaded by the manager
        :return: list<DataWidget>
        """

        return self.standard_data_classes + self.load_data_classes(_reload=_reload)

    def add_directory(self, directory, do_update=False):
        """
        Adds a new directory where data should be find
        :param directory: str
        :param do_update: bool
        """

        if directory not in self.directories:
            self.directories.append(directory)
            if do_update:
                self.load_data_classes()

    def set_directories(self, directories):
        """
        Sets the directories where data should be find
        :param directories: list(str)
        """

        new_dir = False
        for d in directories:
            if d not in self.directories:
                new_dir = True
                self.directories.append(d)

        if new_dir:
            self.load_data_classes()

    def load_data_classes(self, _reload=False):
        """
        Loads all data classes
        :param _reload: bool
        :return: list
        """

        for d in self.directories:
            loaded_classes = self._load_data_classes(directory=d, _reload=_reload)
            self.loaded_data_classes.extend(loaded_classes)

        return self.loaded_data_classes

    def get_available_types(self):
        """
        Returns a list with all available data types
        :return: list<str>
        """

        data_types = list()
        for data in self.get_all_data_classes():
            data_types.append(data.get_data_type())

        return data_types

    def get_type_instance(self, data_type):
        """
        Returns a new instance of data type
        :param data_type: str, type of data type instance we want to create
        :return: variant
        """

        for data in self.get_all_data_classes():
            if data.is_type_match(data_type):
                return data()

    def _load_data_classes(self, directory, _reload=False):

        imported = list()

        if directory is None or not os.path.exists(directory):
            tpRigToolkit.logger.warning('Data Path {} does not exists!'.format(directory))
            return imported

        for loader, mod_name, is_package in pkgutil.walk_packages([directory]):
            try:
                module = loader.find_module(mod_name).load_module(mod_name)
                if _reload:
                    reload(module)
                for cname, obj in inspect.getmembers(module, inspect.isclass):
                    if not issubclass(obj, core_data.FileData):
                        continue
                    # globals()[cname] = obj
                    imported.append(obj)
            except Exception as e:
                tpRigToolkit.logger.warning('Aborting loading Data Class {} : {}'.format(mod_name, str(e)))
                tpRigToolkit.logger.debug(traceback.format_exc())

        return sorted(list(set(imported)))


@decorators.Singleton
class ScriptsManagerSingleton(ScriptsManager, object):
    def __init__(self):
        ScriptsManager.__init__(self)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains implementation for names manager
"""

from __future__ import print_function, division, absolute_import

import tpDcc
from tpDcc.libs.python import python, decorators
from tpDcc.libs.nameit.core import namelib

import tpRigToolkit


@decorators.Singleton
class RigToolkitNameLib(namelib.NameLib, object):
    def __init__(self, naming_file=None, dev=False):
        namelib.NameLib.__init__(self)
        environment = 'development' if dev else 'production'
        config = tpDcc.ConfigsMgr().get_config('tpRigToolkit-names', environment=environment)
        self.naming_file = naming_file or config.get_path()
        self.init_naming_data()


class RigToolkitNamesManager(object):

    def get_auto_suffixes(self, dev=True):
        """
        Returns dictionary containing suffixes that can be used to handle nomenclature
        :return:
        """

        environment = 'development' if dev else 'production'
        naming_config = tpDcc.ConfigsMgr().get_config(
            config_name='tpRigToolkit-naming', environment=environment)
        if not naming_config:
            return dict()

        auto_suffixes_dict = naming_config.get('auto_suffixes', default=dict())
        if not auto_suffixes_dict:
            return dict()

        return auto_suffixes_dict

    def parse_name(self, node_name, rule_name=None, naming_file=None, dev=False):
        """
        Parse a current solved name and return its different fields (metadata information)
        :param node_name: str
        :param rule_name: str
        :param project: str
        :param dev: bool
        :return: dict(str)
        """

        name_lib = RigToolkitNameLib(naming_file=naming_file, dev=dev)
        rule_name = rule_name or 'default'
        rule = None
        if name_lib.has_rule(rule_name):
            rule = name_lib.get_rule(rule_name)
        if not rule:
            tpRigToolkit.logger.warning(
                'Impossible to retrieve name because rule name "{}" is not defined!'.format(rule_name))
            return None

        current_rule = name_lib.active_rule()
        name_lib.set_active_rule(rule_name)
        parsed_name = name_lib.parse(node_name)
        if current_rule:
            if rule_name != current_rule.name:
                name_lib.set_active_rule(current_rule.name)
        else:
            name_lib.set_active_rule(None)

        return parsed_name

    def solve_node_name_by_type(self, node_names=None, naming_file=None, dev=False, **kwargs):
        """
        Resolves node name taking into account its type
        In this case, the type of the node will be used to retrieve an an automatic rule
        The rule name will be retrieved using auto_suffix dict from tpDcc-naming configuration file
        :param node_names: str or list, name of the node we want to take name of
        :param naming_file: str
        :param dev: bool
        :return: str
        """

        if not tpDcc.is_maya():
            tpRigToolkit.logger.warning('Solve Node Name by type functionality is only supported in Maya!')
            return None

        import tpDcc.dccs.maya as maya
        from tpDcc.dccs.maya.core import name

        name_lib = RigToolkitNameLib(naming_file=naming_file, dev=dev)

        auto_suffix = self.get_auto_suffixes()
        if not auto_suffix:
            tpRigToolkit.logger.warning(
                'Impossible to launch auto suffix functionality because no auto suffixes are defined!')
            return None

        solved_names = dict()
        return_names = list()

        if not node_names:
            node_names = tpDcc.Dcc.selected_nodes()
        if not node_names:
            return
        node_names = python.force_list(node_names)

        for obj_name in node_names:
            obj_uuid = maya.cmds.ls(obj_name, uuid=True)
            if not obj_uuid:
                continue
            obj_uuid = obj_uuid[0]
            if obj_uuid in solved_names:
                tpRigToolkit.logger.warning(
                    'Node with name: "{} and UUID "{}" already renamed to "{}"! Skipping ...'.format(
                        obj_name, obj_uuid, solved_names[obj_name]))
                continue

            # TODO: This code is a duplicated version of the one in
            #  tpDcc.dccs.maya.core.name.auto_suffix_object function. Move this code to a DCC specific function
            obj_type = maya.cmds.objectType(obj_name)
            if obj_type == 'transform':
                shape_nodes = maya.cmds.listRelatives(obj_name, shapes=True, fullPath=True)
                if not shape_nodes:
                    obj_type = 'group'
                else:
                    obj_type = maya.cmds.objectType(shape_nodes[0])
            elif obj_type == 'joint':
                shape_nodes = maya.cmds.listRelatives(obj_name, shapes=True, fullPath=True)
                if shape_nodes and maya.cmds.objectType(shape_nodes[0]) == 'nurbsCurve':
                    obj_type = 'controller'
            if obj_type == 'nurbsCurve':
                connections = maya.cmds.listConnections('{}.message'.format(obj_name))
                if connections:
                    for node in connections:
                        if maya.cmds.nodeType(node) == 'controller':
                            obj_type = 'controller'
                            break
            if obj_type not in auto_suffix:
                rule_name = 'node'
                node_type = obj_type
            else:
                rule_name = auto_suffix[obj_type]
                node_type = auto_suffix[obj_type]

            if not name_lib.has_rule(rule_name):
                if not name_lib.has_rule('node'):
                    tpRigToolkit.logger.warning(
                        'Impossible to rename node "{}" by its type "{}" because rule "{}" it is not defined and '
                        'callback rule "node" either'.format(obj_name, obj_type, rule_name))
                else:
                    rule_name = 'node'

            if rule_name == 'node':
                solved_name = self.solve_name(obj_name, node_type=node_type, rule_name=rule_name, **kwargs)
            else:
                solved_name = self.solve_name(rule_name, obj_name, rule_name=rule_name, **kwargs)
            solved_names[obj_uuid] = solved_name

        if not solved_names:
            return

        unique_name = kwargs.get('unique_name', False)

        for obj_id, solved_name in solved_names.items():
            obj_name = maya.cmds.ls(obj_id, long=True)[0]
            if not solved_names:
                return_names.append(obj_name)
            else:
                if unique_name:
                    solved_name = tpDcc.Dcc.find_unique_name(solved_name)
                new_name = name.rename(obj_name, solved_name, uuid=obj_id, rename_shape=True)
                return_names.append(new_name)

        return return_names

    def solve_name(self, *args, **kwargs):
        """
        Resolves name with given rule and attributes
        :param args: list
        :param kwargs: dict
        """

        naming_file = kwargs.get('naming_file', None)
        dev = kwargs.get('dev', False)
        use_auto_suffix = kwargs.pop('use_auto_suffix', True)
        node_type = kwargs.get('node_type', None)
        rule_name = kwargs.pop('rule_name', 'default')

        if use_auto_suffix and node_type:
            auto_suffixes = tpRigToolkit.NamesMgr().get_auto_suffixes() or dict()
            if node_type in auto_suffixes:
                kwargs['node_type'] = auto_suffixes[node_type]

        name_lib = RigToolkitNameLib(naming_file=naming_file, dev=dev)

        fallback_default = False
        if name_lib.has_rule(rule_name):
            rule = name_lib.get_rule(rule_name)
        else:
            tpRigToolkit.logger.warning('Rule with name "{}" is not defined. Default rule used'.format(rule_name))
            rule = name_lib.get_rule('default')
            fallback_default = True

        if not rule:
            tpRigToolkit.logger.warning(
                'Impossible to retrieve name because rule name "{}" is not defined!'.format(rule_name))
            return None

        current_rule = name_lib.active_rule()
        name_lib.set_active_rule(rule_name)
        solved_name = name_lib.solve(*args, **kwargs)
        if current_rule:
            if rule_name != current_rule.name:
                name_lib.set_active_rule(current_rule.name)
        else:
            name_lib.set_active_rule(None)

        solved_name = solved_name or kwargs.get('default', None)

        unique_name = kwargs.get('unique_name', False)
        if solved_name and unique_name:
            solved_name = tpDcc.Dcc.find_unique_name(solved_name)

        if fallback_default:
            tpRigToolkit.logger.warning('Solved name with no rule (used default one): {}'.format(solved_name))

        return solved_name


@decorators.Singleton
class RigToolkitNamesManagerSingleton(RigToolkitNamesManager, object):
    def __init__(self):
        RigToolkitNamesManager.__init__(self)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains options to select rig controls
"""

from __future__ import print_function, division, absolute_import

import os

from Qt.QtCore import *
from Qt.QtWidgets import *

import tpDcc as tp
from tpDcc.libs.qt.core import base
from tpDcc.libs.qt.widgets import layouts, label, lineedit, buttons

from tpRigToolkit.widgets.options import rigoption


class RigControlOption(rigoption.RigOption, object):
    def __init__(self, name, parent, main_widget, rig_object):
        super(RigControlOption, self).__init__(name=name, parent=parent, main_widget=main_widget, rig_object=rig_object)

    def get_option_type(self):
        return 'rigcontrol'

    def get_option_widget(self):
        return GetControlRigWidget(name=self._name, rig_object=self._rig)

    def get_name(self):
        name = self._option_widget.get_name()
        return name

    def set_name(self, name):
        self._option_widget.set_name(name)

    def get_value(self):
        return self._option_widget.control_data

    def set_value(self, value):
        self._option_widget.set_value(value)

    def _setup_option_widget_value_change(self):
        self._option_widget.valueChanged.connect(self._on_value_changed)


class GetControlRigWidget(base.BaseWidget, object):
    valueChanged = Signal(object)

    def __init__(self, name, rig_object, parent=None):
        self._name = name
        self._rig = rig_object
        super(GetControlRigWidget, self).__init__(parent=parent)

        self._control_data = dict()

    @property
    def control_data(self):
        return self._control_data

    def get_main_layout(self):
        main_layout = layouts.HorizontalLayout()
        main_layout.setSpacing(2)
        main_layout.setContentsMargins(2, 2, 2, 2)

        return main_layout

    def ui(self):
        super(GetControlRigWidget, self).ui()

        self._label = label.BaseLabel(self._name)
        self._label.setAttribute(Qt.WA_TransparentForMouseEvents)
        self._control = ControlLineEdit(self._rig)

        self.main_layout.addWidget(self._label)
        self.main_layout.addWidget(self._control)

    def setup_signals(self):
        self._control.controlSelected.connect(self._on_selected_control)

    def get_value(self):
        return self._control_data

    def set_value(self, value_dict):
        self._control_data = value_dict or dict()
        self._control.set_data(self._control_data)

    def get_name(self):
        return self._label.text()

    def set_name(self, value):
        self._label.setText(value)

    def _on_selected_control(self, control_data):
        self.set_value(control_data)
        self.valueChanged.emit(control_data)


class ControlLineEdit(base.BaseWidget, object):
    controlSelected = Signal(object)

    def __init__(self, rig_object, parent=None):
        self._rig_object = rig_object
        self._control_data = dict()
        super(ControlLineEdit, self).__init__(parent=parent)

    @property
    def control_data(self):
        return self._control_data

    def get_main_layout(self):
        main_layout = layouts.HorizontalLayout()
        main_layout.setSpacing(2)
        main_layout.setContentsMargins(2, 2, 2, 2)

        return main_layout

    def ui(self):
        super(ControlLineEdit, self).ui()

        self._line = lineedit.BaseLineEdit()
        self._btn = buttons.BaseButton(text='...')

        self.main_layout.addWidget(self._line)
        self.main_layout.addWidget(self._btn)

    def setup_signals(self):
        self._btn.clicked.connect(self._on_open_rig_control_selector)

    def set_data(self, data):
        data = data if data is not None else dict()
        name = data.get('control_name', '')
        self._line.setText(str(name))
        self._line.setToolTip(str(data))
        self._control_data = data

    def _on_open_rig_control_selector(self):

        from tpRigToolkit.tools.controlrig.widgets import controlrig

        dlg = QDialog(parent=tp.Dcc.get_main_window() or None)
        dlg.setWindowTitle('Select Control')
        lyt = QVBoxLayout()
        lyt.setSpacing(0)
        lyt.setContentsMargins(0, 0, 0, 0)
        dlg.setLayout(lyt)

        if hasattr(self._rig_object, 'full_path'):
            current_project = self._rig_object
        else:
            current_project = self._rig_object.project
        project_path = current_project.full_path

        # TODO: The project itself should return this path
        controls_path = os.path.join(project_path, 'controls.json')

        control_selector = controlrig.ControlSelector(controls_path=controls_path)
        if self._control_data:
            control_selector.set_control_data(self._control_data)
        lyt.addWidget(control_selector)
        dlg.resize(600, 700)
        dlg.exec_()
        control_data = control_selector.control_data or dict()
        if not control_data and self._control_data:
            return
        if control_data:
            control_data.pop('shape_data', None)
            control_data.pop('name', None)

        self.controlSelected.emit(control_data)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains classes to create editor tools inside Qt apps
"""

from __future__ import print_function, division, absolute_import

import uuid

from Qt.QtCore import *
from Qt.QtWidgets import *

import tpDcc
import tpRigToolkit


class BasePlugin(object):
    """
    Base class used by all  editor tools
    """

    SUPPORTED_SOFTWARES = ['any']
    TOOLTIP = 'Default Tooltip'
    NAME = 'BasePlugin'

    def __init__(self):
        super(BasePlugin, self).__init__()

        self._uid = uuid.uuid4()
        self._app = None

    @property
    def uid(self):
        """
        Returns unique identifier of the tool
        :return: uuid
        """

        return self._uid

    @property
    def app(self):
        """
        Returns app that manages this tool
        :return:
        """

        return self._app

    @app.setter
    def app(self, tool_app):
        """
        Sets the app that manages this tool
        :param tool_app:
        """

        self._app = tool_app

    @staticmethod
    def icon():
        """
        Returns the icon of the tool
        :return: QIcon or None
        """

        return None

    def unique_name(self):
        """
        Returns unique name of the tool
        When a tool is not singleton, we need to store separate data for each instance.
        We use unique identifier for that
        :return: str
        """

        return '{}::{}'.format(self.NAME, str(self.uid))

    def show_plugin(self):
        """
        Internal function that is called when the tool pops up
        """

        pass

    def save_state(self, settings):
        """
        Saves the settings of the tool in the given settings
        NOTE: Tool Settings group is already selected when this function is called, so there is no need
        to open/end tool settings group in this function
        :param settings: QtSettings
        """

        settings.setValue('uid', str(self.uid))

    def restore_state(self, settings):
        """
        Restore any save state from given settings
        NOTE: Tool Settings group is already selected when this function is called, so there is no need
        to open/end tool settings group in this function
        :param settings: QtSettings
        """

        uid_str = settings.value(str(self.uid))
        if uid_str:
            self._uid = uuid.UUID(uid_str)
        else:
            self._uid = uuid.uuid4()

    def close_plugin(self):
        """
        Function that is called when tol is closed
        """

        pass


class ShelfPlugin(BasePlugin, object):
    def __init__(self):
        super(ShelfPlugin, self).__init__()

    @staticmethod
    def icon():
        return tpDcc.ResourcesMgr().icon('home')

    def context_menu_builder(self):
        return None

    def do(self):
        print(self.NAME, 'called!')


class DockPlugin(QDockWidget, BasePlugin):

    closed = Signal(object)

    DEFAULT_DOCK_AREA = Qt.LeftDockWidgetArea
    IS_SINGLETON = False

    def __init__(self):
        BasePlugin.__init__(self)
        QDockWidget.__init__(self)

        self.setToolTip(self.TOOLTIP)
        self.setFeatures(QDockWidget.AllDockWidgetFeatures)
        self.setAllowedAreas(
            Qt.BottomDockWidgetArea | Qt.LeftDockWidgetArea | Qt.RightDockWidgetArea | Qt.TopDockWidgetArea)
        self.setObjectName(self.unique_name())
        self.setTitleBarWidget(DockTitleBar(self))
        self.setFloating(False)

    def restore_state(self, settings):
        super(DockPlugin, self).restore_state(settings)

        self.setObjectName(self.unique_name())

    def closeEvent(self, event):
        """
        Overrides base QDockWidget closeEvent function
        :param event: QEvent
        """

        self.close_plugin()
        tpRigToolkit.PluginsMgr().unregister_plugin_instance(self)
        self.closed.emit(self)
        event.accept()

    def add_button(self, button):
        self.titleBarWidget().add_button(button)

    def show_plugin(self):
        """
        Overrides base BaseTool _on_show function
        """

        super(DockPlugin, self).show_plugin()
        self.setWindowTitle(self.NAME)
        self.show()


class DockTitleBar(QWidget, object):
    def __init__(self, dock_widget, renamable=False):
        super(DockTitleBar, self).__init__(dock_widget)

        self._renamable = renamable

        main_layout = QHBoxLayout()
        main_layout.setContentsMargins(0, 0, 0, 1)
        self.setLayout(main_layout)

        self._buttons_box = QGroupBox('')
        self._buttons_box.setObjectName('Docked')
        self._buttons_layout = QHBoxLayout()
        self._buttons_layout.setSpacing(1)
        self._buttons_layout.setMargin(2)
        self._buttons_box.setLayout(self._buttons_layout)
        main_layout.addWidget(self._buttons_box)

        self._title_label = QLabel(self)
        self._title_label.setStyleSheet('background: transparent')
        self._title_edit = QLineEdit(self)
        self._title_edit.setVisible(False)

        self._button_size = QSize(14, 14)

        self._dock_btn = QToolButton(self)
        self._dock_btn.setIcon(tpDcc.ResourcesMgr().icon('restore_window', theme='color'))
        self._dock_btn.setMaximumSize(self._button_size)
        self._dock_btn.setAutoRaise(True)
        self._close_btn = QToolButton(self)
        self._close_btn.setIcon(tpDcc.ResourcesMgr().icon('close_window', theme='color'))
        self._close_btn.setMaximumSize(self._button_size)
        self._close_btn.setAutoRaise(True)

        self._buttons_layout.addSpacing(2)
        self._buttons_layout.addWidget(self._title_label)
        self._buttons_layout.addWidget(self._title_edit)
        self._buttons_layout.addStretch()
        self._buttons_layout.addSpacing(5)
        self._buttons_layout.addWidget(self._dock_btn)
        self._buttons_layout.addWidget(self._close_btn)

        self._buttons_box.mouseDoubleClickEvent = self.mouseDoubleClickEvent
        self._buttons_box.mousePressEvent = self.mousePressEvent
        self._buttons_box.mouseMoveEvent = self.mouseMoveEvent
        self._buttons_box.mouseReleaseEvent = self.mouseReleaseEvent

        dock_widget.featuresChanged.connect(self._on_dock_features_changed)
        self._title_edit.editingFinished.connect(self._on_finish_edit)
        self._dock_btn.clicked.connect(self._on_dock_btn_clicked)
        self._close_btn.clicked.connect(self._on_close_btn_clicked)

        self._on_dock_features_changed(dock_widget.features())
        self.set_title(dock_widget.windowTitle())
        dock_widget.installEventFilter(self)
        dock_widget.topLevelChanged.connect(self._on_change_floating_style)

    @property
    def renamable(self):
        return self._renamable

    @renamable.setter
    def renamable(self, flag):
        self._renamable = flag

    def eventFilter(self, obj, event):
        try:
            if event.type() == QEvent.WindowTitleChange:
                self.set_title(obj.windowTitle())
            return super(DockTitleBar, self).eventFilter(obj, event)
        except Exception as exc:
            event.accept()
            return True

    def mouseMoveEvent(self, event):
        event.ignore()

    def mousePressEvent(self, event):
        event.ignore()

    def mouseReleaseEvent(self, event):
        event.ignore()

    def mouseDoubleClickEvent(self, event):
        if event.pos().x() <= self._title_label.width() and self._renamable:
            self._start_edit()
        else:
            super(DockTitleBar, self).mouseDoubleClickEvent(event)

    def update(self, *args, **kwargs):
        self._on_change_floating_style(self.parent().isFloating())
        super(DockTitleBar, self).update(*args, **kwargs)

    def set_title(self, title):
        self._title_label.setText(title)
        self._title_edit.setText(title)

    def add_button(self, button):
        button.setAutoRaise(True)
        button.setMaximumSize(self._button_size)
        self._buttons_layout.insertWidget(5, button)

    def _start_edit(self):
        self._title_label.hide()
        self._title_edit.show()
        self._title_edit.setFocus()

    def _finish_edit(self):
        self._title_edit.hide()
        self._title_label.show()
        self.parent().setWindowTitle(self._title_edit.text())

    def _on_dock_features_changed(self, features):
        if not features & QDockWidget.DockWidgetVerticalTitleBar:
            self._close_btn.setVisible(features & QDockWidget.DockWidgetClosable)
            self._dock_btn.setVisible(features & QDockWidget.DockWidgetFloatable)
        else:
            raise ValueError('Vertical title bar is not supported!')

    def _on_finish_edit(self):
        self._finish_edit()

    def _on_dock_btn_clicked(self):
        self.parent().setFloating(not self.parent().isFloating())

    def _on_close_btn_clicked(self):
        self.parent().toggleViewAction().setChecked(False)
        self.parent().close()

    def _on_change_floating_style(self, state):
        pass


class PluginContextMenuGenerator(object):
    def __init__(self, menu_data_builder):
        super(PluginContextMenuGenerator, self).__init__()
        self._builder = menu_data_builder

    def build(self):
        menu_data = self._builder.get()
        menu = QMenu()
        for menu_entry in menu_data:
            self._create_menu_entry(menu, menu_entry)

        return menu

    def _create_menu_entry(self, parent_menu, menu_entry_data):
        if 'separator' in menu_entry_data:
            parent_menu.addSeparator()
            return
        icon = menu_entry_data['icon']
        if 'sub_menu' in menu_entry_data:
            sub_menu_data = menu_entry_data['sub_men']
            sub_menu = parent_menu.addMenu(menu_entry_data['title'])
            if icon:
                sub_menu.setIcon(icon)
            self._create_menu_entry(sub_menu, sub_menu_data)
        else:
            action = parent_menu.addAction(menu_entry_data['title'])
            if icon:
                action.setIcon(icon)
            action.triggered.connect(menu_entry_data['callback'])


def create_plugin_instance(plugin_class, already_registered_plugins=None):
    """
    Creates a tool instance of the given class
    :param plugin_class: cls
    :param already_registered_plugins: list
    :return:
    """

    import tpDcc as tp

    if already_registered_plugins is None:
        already_registered_plugins = list()

    if not plugin_class:
        return

    is_singleton = plugin_class.IS_SINGLETON
    if is_singleton:
        if plugin_class.NAME in [t.NAME for t in already_registered_plugins]:
            for registered_plugin in already_registered_plugins:
                if registered_plugin.NAME == plugin_class.NAME:
                    registered_plugin.show()
                    registered_plugin.show_plugin()
                    return registered_plugin
            return None

    supported_softwares = plugin_class.SUPPORTED_SOFTWARES
    if 'any' not in supported_softwares:
        if tp.Dcc.get_name() not in supported_softwares:
            tpDcc.logger.warning(
                'Plugin {} is not suppported in current software: "{}"'.format(plugin_class.NAME, tp.Dcc.get_name()))
            return None

    return plugin_class()



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains renamer tool implementation
"""

from __future__ import print_function, division, absolute_import

from Qt.QtCore import *
from Qt.QtWidgets import *

from tpRigToolkit.tools.renamer.widgets import renamer

from tpRigToolkit.core import plugin


class RenamerPlugin(plugin.DockPlugin, object):

    NAME = 'Renamer'
    TOOLTIP = 'Allows to rename DCC nodes'
    DEFAULT_DOCK_AREA = Qt.LeftDockWidgetArea

    def __init__(self):
        super(RenamerPlugin, self).__init__()

        self._renamer_widget = None
        self._content = QWidget()
        self._content_layout = QVBoxLayout()
        self._content_layout.setContentsMargins(0, 0, 0, 0)
        self._content_layout.setSpacing(0)
        self._content.setLayout(self._content_layout)
        self.setWidget(self._content)

    def show_plugin(self):
        super(RenamerPlugin, self).show_plugin()

        self._renamer_widget = RenamerWidget()
        self._renamer_widget.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self._content_layout.addWidget(self._renamer_widget)


class RenamerWidget(renamer.RigToolkitRenamerWidget, object):

    def __init__(self, config=None, parent=None):
        super(RenamerWidget, self).__init__(config=config, parent=parent)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains data manager implementation
"""

from __future__ import print_function, division, absolute_import

import os
import pkgutil
import inspect
import traceback

from tpDcc.core import scripts, data as core_data
from tpDcc.libs.python import path, decorators
from tpDcc.libs.qt.widgets.library import manager

import tpRigToolkit
from tpRigToolkit.core import data, utils


class DataManager(manager.LibraryManager, object):
    def __init__(self, settings=None, directories=None, update_on_init=True):
        super(DataManager, self).__init__(settings=settings)

        data_dirs = utils.get_data_files_directory()
        data_dirs.extend(directories or list())

        self._directories = list(set(data_dirs))
        self._loaded_data_items = list()
        self._loaded_data_classes = list()

        self.standard_data_classes = [
            scripts.ScriptManifestData,
            scripts.ScriptPythonData
        ]

        if update_on_init:
            self.update_data_classes()
            self.update_data_items()

    def get_all_data_classes(self, do_reload=False):
        """
        Returns all data classes loaded by the manager
        :return: list<DataWidget>
        """

        return self.standard_data_classes + self.update_data_classes(do_reload=do_reload)

    def get_all_data_items(self, do_reload=False):
        """
        Returns all data item classes loaded by the manager
        :param do_reload:
        :return:
        """

        return self.update_data_items(do_reload=do_reload)

    def get_available_types(self):
        """
        Returns a list with all available data types
        :return: list<str>
        """

        data_types = list()
        for data_class in self.get_all_data_classes():
            data_types.append(data_class.get_data_type())

        return data_types

    def add_directory(self, directory, do_update=False):
        """
        Adds a new directory where data should be find
        :param directory: str
        :param do_update: bool
        """

        if directory not in self._directories:
            self._directories.append(directory)
            if do_update:
                self.update_data_classes()

    def set_directories(self, directories):
        """
        Sets the directories where data should be find
        :param directories: list(str)
        """

        new_dir = False
        for d in directories:
            if d not in self._directories:
                new_dir = True
                self._directories.append(d)

        if new_dir:
            self.update_data_classes(do_reload=True)
            self.update_data_items(do_reload=True)

    def update_data_classes(self, do_reload=False):
        """

        :param do_reload:
        :return:
        """

        if not self._loaded_data_classes or do_reload:
            for d in self._directories:
                self._loaded_data_classes.extend(self._load_data_classes(d))

        return self._loaded_data_classes

    def update_data_items(self, do_reload=False):
        """
        Adds custom dat files located in the current data manager registered directories
        """

        if not self._loaded_data_items or do_reload:
            for d in self._directories:
                self._loaded_data_items.extend(self._load_data_items(d))
            for data_item in self._loaded_data_items:
                self.register_item(data_item)

        return self._loaded_data_items

    def get_type_instance(self, data_type):
        """
        Returns a new instance of data type
        :param data_type: str, type of data type instance we want to create
        :return: variant
        """

        for data_class in self.get_all_data_classes():
            if data_class.is_type_match(data_type):
                return data_class()

    def _load_data_classes(self, directory):

        imported = list()

        if directory is None or not os.path.exists(directory):
            tpRigToolkit.logger.warning('Data Path {} does not exists!'.format(directory))
            return imported

        for loader, mod_name, is_package in pkgutil.walk_packages([directory]):
            try:
                module = loader.find_module(mod_name).load_module(mod_name)
                for cname, obj in inspect.getmembers(module, inspect.isclass):
                    if not issubclass(obj, core_data.FileData):
                        continue
                    # globals()[cname] = obj
                    imported.append(obj)
            except Exception as e:
                tpRigToolkit.logger.warning('Aborting loading Data Class {} : {}'.format(mod_name, str(e)))
                tpRigToolkit.logger.debug(traceback.format_exc())

        # TODO: Not working in Python 3
        # return sorted(list(set(imported)))

        return list(set(imported))

    def _load_data_items(self, directory):
        """
        Internal function that loads data classes
        :param directory: str
        :return: list
        """

        data_classes = list()

        if directory is None or not path.is_dir(directory):
            tpRigToolkit.logger.warning('Data Path {} does not exists!'.format(directory))
            return data_classes

        for loader, mod_name, is_package in pkgutil.walk_packages([directory]):
            try:
                module = loader.find_module(mod_name).load_module(mod_name)
                for cname, obj in inspect.getmembers(module, inspect.isclass):
                    if not issubclass(obj, data.DataItem):
                        continue
                    data_classes.append(obj)
            except Exception as e:
                tpRigToolkit.logger.warning('Aborting loading Data Class {} : {}'.format(mod_name, str(e)))
                tpRigToolkit.logger.error(traceback.format_exc())

        for data_cls in data_classes:
            tpRigToolkit.logger.info('Found Data Class: {}'.format(data_cls))

        # TODO: Not working in Python 3
        # '<' not supported between instances of 'Shiboken.ObjectType' and 'Shiboken.ObjectType'
        # return sorted(list(set(data_classes)))
        return list(set(data_classes))


@decorators.Singleton
class DataManagerSingleton(DataManager, object):
    def __init__(self):
        DataManager.__init__(self)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains factory class to create tpRigToolkit options
"""

from tpDcc.libs.python import python, decorators
from tpDcc.libs.qt.widgets.options import factory

from tpRigToolkit.widgets.options import bone, control, bonecontrollink


class RigOptionsFactory(factory.OptionsFactory, object):
    def __init__(self):
        super(RigOptionsFactory, self).__init__()

    def add_option(self, option_type, name=None, value=None, parent=None, main_widget=None, option_object=None):
        if option_type == 'rigcontrol':
            new_option = self._add_control_rig(
                name=name, value=value, parent=parent, main_widget=main_widget, option_object=option_object)
        elif option_type == 'bone':
            new_option = self._add_bone(name=name, value=value, parent=parent, main_widget=main_widget)
        elif option_type == 'boneList':
            new_option = self._add_bone_list(name=name, value=value, parent=parent, main_widget=main_widget)
        elif option_type == 'boneControlLink':
            new_option = self._add_bone_control_link(
                name=name, value=value, parent=parent, main_widget=main_widget, option_object=option_object)
        else:
            new_option = super(RigOptionsFactory, self).add_option(
                option_type, name, value, parent, main_widget, option_object)

        return new_option

    def _add_control_rig(self, name='rigcontrol', value=None, parent=None, main_widget=None, option_object=None):
        if type(name) == bool:
            name = 'rigcontrol'

        control_option = control.RigControlOption(
            name=name, parent=parent, main_widget=main_widget, rig_object=option_object)
        control_option.set_value(value)

        return control_option

    def _add_bone(self, name='bone', value=None, parent=None, main_widget=None):
        if type(name) == bool:
            name = 'bone'

        bone_option = bone.BoneOption(name=name, parent=parent, main_widget=main_widget)
        bone_option.set_value(value)

        return bone_option

    def _add_bone_list(self, name='boneList', value=None, parent=None, main_widget=None):
        if type(name) == bool:
            name = 'boneList'

        value = python.force_list(value)
        bone_list_option = bone.BoneOptionList(name=name, parent=parent, main_widget=main_widget)
        bone_list_option.set_value(value)

        return bone_list_option

    def _add_bone_control_link(
            self, name='boneControlLink', value=None, parent=None, main_widget=None, option_object=None):
        if type(name) == bool:
            name = 'boneControlLink'

        bone_control_link_option = bonecontrollink.BoneControlLinkOption(
            name=name, parent=parent, main_widget=main_widget, rig_object=option_object)
        bone_control_link_option.set_value(value or list())

        return bone_control_link_option


@decorators.Singleton
class RigOptionsFactorySingleton(RigOptionsFactory, object):
    def __init__(self):
        RigOptionsFactory.__init__(self)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains manager that handles the registration of new tpRigToolkit plugins
"""

from __future__ import print_function, division, absolute_import

import tpDcc as tp
from tpDcc.libs.python import decorators

from tpRigToolkit.core import log, plugin


class PluginsManager(object):

    def __init__(self):
        super(PluginsManager, self).__init__()

        self._plugin_classes = list()
        self._plugins = set()

    # ============================================================================================================
    # PROPERTIES
    # ============================================================================================================

    @property
    def plugin_classes(self):
        return self._plugin_classes

    # ============================================================================================================
    # BASE
    # ============================================================================================================

    def get_registered_plugins(self, class_name_filters=None):
        """
        Returns a list with all registered plugin instances
        :param class_name_filters: str, String to filter plugins to search for
        :return: list(str)
        """

        if class_name_filters is None:
            class_name_filters = list()

        if len(class_name_filters) == 0:
            return self._plugins
        else:
            result = list()
            for plugin_inst in self._plugins:
                if plugin_inst.__class__.__name__ in class_name_filters:
                    result.append(plugin_inst)

            return result

    def is_plugin_opened(self, plugin_name):
        """
        Returns whether or not a plugin with given name is already opened
        :param plugin_name: str
        :return: bool
        """

        return plugin_name in [t.NAME for t in self._plugins]

    def get_plugin_instance(self, plugin_name):
        """
        Returns plugin instance of the given plugin
        :param plugin_name: str
        :return: list(object)
        """

        plugins_found = list()
        for plugin_instance in self._plugins:
            if plugin_instance.NAME == plugin_name:
                plugins_found.append(plugin_instance)

        return plugins_found

    def register_plugin_class(self, plugin_class):
        """
        Registers given tool class
        :param plugin_class: cls
        """

        if not plugin_class or plugin_class in self._plugin_classes:
            return

        self._plugin_classes.append(plugin_class)

    def invoke_dock_plugin_by_name(self, plugin_name, parent_window=None, settings=None):
        plugin_class = None
        parent_window = parent_window or tp.Dcc.get_main_window()
        for t in self._plugin_classes:
            if t.NAME == plugin_name:
                plugin_class = t
                break
        if not plugin_class:
            log.warning('No registered tool found with name: "{}"'.format(plugin_name))
            return None

        tool_instance = plugin.create_plugin_instance(plugin_class, self._plugins)
        if not tool_instance:
            return None
        if plugin_class.NAME in [t.NAME for t in self._plugins] and plugin_class.IS_SINGLETON:
            return tool_instance

        self.register_plugin_instance(tool_instance)
        if settings:
            tool_instance.restore_state(settings)
            if not self.restoreDockWidget(tool_instance):
                pass
        else:
            parent_window.addDockWidget(tool_instance.DEFAULT_DOCK_AREA, tool_instance)

        tool_instance.app = parent_window
        tool_instance.show_plugin()

        return tool_instance

    def register_plugin_instance(self, instance):
        """
        Internal function that registers given plugin instance
        Used to prevent plugin classes being garbage collected and to save plugin widgets states
        :param instance: Tool
        """

        self._plugins.add(instance)

    def unregister_plugin_instance(self, instance):
        """
        Internal function that unregister plugin instance
        :param instance: Tool
        """

        if instance not in self._plugins:
            return False
        self._plugins.remove(instance)

        return True


@decorators.Singleton
class PluginsManagerSingleton(PluginsManager, object):
    def __init__(self):
        PluginsManager.__init__(self)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Initialization module for tpRigToolkit-dccs
"""

from __future__ import print_function, division, absolute_import

__author__ = "Tomas Poveda"
__license__ = "MIT"
__maintainer__ = "Tomas Poveda"
__email__ = "tpovedatd@gmail.com"

from pkgutil import extend_path

__path__ = extend_path(__path__, __name__)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Register module for tpRigToolkit
"""

from __future__ import print_function, division, absolute_import

__author__ = "Tomas Poveda"
__license__ = "MIT"
__maintainer__ = "Tomas Poveda"
__email__ = "tpovedatd@gmail.com"

import tpRigToolkit


# =================================================================================

REGISTER_ATTR = '_registered_classes'

# =================================================================================


def register_class(cls_name, cls, is_unique=True, skip_store=False):
    """
    This function registers given class into tpRigToolkit module
    :param cls_name: str, name of the class we want to register
    :param cls: class, class we want to register
    :param is_unique: bool, Whether if the class should be updated if new class is registered with the same name
    :param skip_store: bool, Whether the registered class should be removed during cleanup operation
        Useful in scenarios where we want to cleanup registered class manually.
    """

    if REGISTER_ATTR not in tpRigToolkit.__dict__:
        tpRigToolkit.__dict__[REGISTER_ATTR] = list()

    if not is_unique and cls_name in tpRigToolkit.__dict__:
        return

    tpRigToolkit.__dict__[cls_name] = cls
    if not skip_store:
        tpRigToolkit.__dict__[REGISTER_ATTR].append(cls_name)


def cleanup():

    if REGISTER_ATTR not in tpRigToolkit.__dict__:
        return

    for cls_name in tpRigToolkit.__dict__[REGISTER_ATTR]:
        if cls_name not in tpRigToolkit.__dict__:
            continue
        # print('Deleting: {}'.format(cls_name))
        del tpRigToolkit.__dict__[cls_name]
    del tpRigToolkit.__dict__[REGISTER_ATTR]



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains functions to interact with tpRigToolkit
"""

from __future__ import print_function, division, absolute_import

import tpDcc as tp
from tpDcc.core import project
from tpDcc.libs.qt.widgets import project

import tpRigToolkit
from tpRigToolkit.core import consts


def get_project_by_name(projects_path, project_name, project_class):
    """
    Returns a project located in the given path and with the given name (if exists)
    :param projects_path: str
    :param project_name: str
    :param project_class: cls
    :return: Project or None
    """

    return project.get_project_by_name(projects_path, project_name, project_class=project_class)


def get_projects(projects_path, project_class):
    """
    Returns all projects located in given path
    :param projects_path: str
    :param project_class: cls
    :return: list(Project)
    """

    return project.get_projects(projects_path, project_class=project_class)


def solve_name(*args, **kwargs):
    """
    Resolves name with given rule and attributes
    :param args: list
    :param kwargs: str
    """

    kwargs['unique_name'] = kwargs.get('unique_name', True)

    solved_name = tpRigToolkit.NamesMgr().solve_name(*args, **kwargs)

    return solved_name


def parse_name(node_name, rule_name=None):
    """
    Parse a current solved name and return its different fields (metadata information)
    :param node_name: str
    :param rule_name: str
    :return: dict(str)
    """

    return tpRigToolkit.NamesMgr().parse_name(node_name=node_name, rule_name=rule_name)


def get_sides(current_project, skip_default=False, default_sides=None, default_side=None):
    """
    Returns sides being used
    This is the order that is used to check the list of available sides
        1) Check if project has already a dict option called sides. Default side will be the first side in the list.
        2) Check project nomenclature rule looking for a rule called side
        3) Default sides for tpRigToolkit will be used
    :return: List of sides and default one
    :rtype: list(str), str
    """

    if current_project:
        # Check project options
        if current_project.has_option('sides'):
            sides = current_project.get_option('sides')
            if sides:
                default_side = sides[0]
                if skip_default:
                    sides.remove(default_side)
                return sides, default_side
        # Check project nomenclature
        name_lib = current_project.naming_lib
        side_token = name_lib.get_token('side')
        if side_token:
            token_items = side_token.get_items().keys()
            token_default = side_token.default or 1
            token_default_value = token_items[token_default - 1] if token_items else ''
            if skip_default:
                if token_default_value in token_items:
                    token_items.remove(token_default_value)
            return token_items, token_default_value
    else:
        # Default fallback
        return default_sides or consts.DEFAULT_SIDES.keys(), default_side or consts.DEFAULT_SIDE


def get_side_long_name(short_side, current_project=None):
    """
    Returns full version of the given side
    :param short_side: str
    :return: str
    """

    if current_project:
        # Check project options
        if current_project.has_option('sides'):
            sides = current_project.get_option('sides')
            if sides:
                for side_name, side_short in sides.items():
                    if side_short.lower() == short_side.lower():
                        return side_name

        # Check project nomenclature
        name_lib = current_project.naming_lib
        side_token = name_lib.get_token('side')
        for side_key, side_value in side_token.get_items().items():
            if side_value.lower() == short_side.lower():
                return side_key
    else:
        # Default fallback
        for side_name, side_short in consts.DEFAULT_SIDE.items():
            if side_short.lower() == short_side.lower():
                return side_name

    return short_side


def get_mirror_side(current_project, default_mirror_side=None):
    """
    Returns current mirror side used by the project
    :return: str
    """

    if current_project:
        # Check project options
        if current_project.has_option('mirror side'):
            return current_project.get_option('mirror side')

        # Check project nomenclature
        name_lib = current_project.naming_lib
        mirror_side_token = name_lib.get_token('mirror side')
        if mirror_side_token:
            token_items = mirror_side_token.get_items().keys()
            token_default = mirror_side_token.default or 1
            token_default_value = token_items[token_default - 1] if token_items else ''
            return token_default_value
        else:
            return default_mirror_side or consts.DEFAULT_MIRROR_SIDE
    else:
        # Default fallback
        return default_mirror_side or consts.DEFAULT_MIRROR_SIDE


def get_default_side(current_project):
    """
    Returns current default side used by the project
    :return: str
    """

    sides, default_side = get_sides(current_project=current_project)

    return default_side


def get_color_of_side(current_project, side, sub_color=False):
    """
    Returns override color of the given side
     This is the order that is used to check the list of available sides
        1) Check if project has already side colors defined.
        2) Default sides for DCC will be used
    :param current_project: str
    :param side: str
    :param sub_color: fool, whether to return a sub color or not
    :return:
    """

    side_color = None
    if current_project:
        groups = ['Controls', 'controls'] if not sub_color else ['SubControls', 'subcontrols']
        for group in groups:
            if current_project.has_option(side, group=group):
                side_color = current_project.get_option(side, group=group)
            else:
                side_long = get_side_long_name(side, current_project=current_project)
                if current_project.has_option(side_long, group=group):
                    side_color = current_project.get_option(side_long, group=group)
                if current_project.has_option(side_long.title(), group=group):
                    side_color = current_project.get_option(side_long.title(), group=group)
            if side_color:
                break

    return side_color if side_color else tp.Dcc.get_color_of_side(side=side, sub_color=sub_color)


def get_side_colors(current_project, sides=None, sub_color=False):
    """
    Returns dictionary containing the side color for each of the side of the current project
    :param current_project: Project
    :param sides: list(str) or None, If not given, default sides will be used
    :param sub_color: bool, Whether to return the color for main controls or sub controls
    :return: dict(str, list)
    """

    side_colors = dict()
    sides, default_side = sides or get_sides(current_project=current_project)
    for side in sides:
        side_color_data = get_color_of_side(current_project, side=side, sub_color=sub_color)
        if side_color_data:
            side_colors[side] = side_color_data

    return side_colors


def get_mirror_name(current_project, node_name):
    """
    Returns the mirrored name of the given node
    :param current_project: str
    :param node_name: str
    :return: str
    """

    all_sides, default_side = get_sides(current_project=current_project, skip_default=True)
    parsed_name = parse_name(node_name)
    if 'side' not in parsed_name or not parsed_name['side'] or parsed_name['side'] not in all_sides:
        return tp.Dcc.get_mirror_name(node_name)
    else:
        mirror_side = parsed_name['side']
        for side in all_sides:
            if side != parsed_name['side']:
                mirror_side = side
                break
        parsed_name['side'] = mirror_side

        return solve_name(parsed_name)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains widget to handle editable options for rigs
"""

from __future__ import print_function, division, absolute_import

from functools import partial

from Qt.QtWidgets import *

import tpDcc as tp
from tpDcc.libs.qt.widgets.options import optionlist

from tpRigToolkit.widgets.options import factory


class RigOptionList(optionlist.OptionList, object):

    FACTORY_CLASS = factory.RigOptionsFactorySingleton

    def __init__(self, parent=None, option_object=None):
        super(RigOptionList, self).__init__(parent=parent, option_object=option_object)

    def _create_context_menu(self, menu, parent=None):
        create_menu = super(RigOptionList, self)._create_context_menu(menu=menu, parent=parent)

        control_icon = tp.ResourcesMgr().icon('rigcontrol')
        bone_icon = tp.ResourcesMgr().icon('bone')
        link_icon = tp.ResourcesMgr().icon('link')

        create_menu.addSeparator()
        add_rig_control_action = QAction(control_icon, 'Add Rig Control', create_menu)
        create_menu.addAction(add_rig_control_action)
        add_bone_action = QAction(bone_icon, 'Add Rig Joint', create_menu)
        create_menu.addAction(add_bone_action)
        add_bone_list_action = QAction(bone_icon, 'Add Rig Joint List', create_menu)
        create_menu.addAction(add_bone_list_action)
        add_control_bone_link_action = QAction(link_icon, 'Add Control/Joint Link', create_menu)
        create_menu.addAction(add_control_bone_link_action)

        add_rig_control_action.triggered.connect(partial(parent._add_option, 'rigcontrol'))
        add_bone_action.triggered.connect(partial(parent._add_option, 'bone'))
        add_bone_list_action.triggered.connect(partial(parent._add_option, 'boneList'))
        add_control_bone_link_action.triggered.connect(partial(parent._add_option, 'boneControlLink'))

        return create_menu



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Base wrapper classes to create DCC windows
"""

from __future__ import print_function, division, absolute_import

__author__ = "Tomas Poveda"
__license__ = "MIT"
__maintainer__ = "Tomas Poveda"
__email__ = "tpovedatd@gmail.com"

import webbrowser

from Qt.QtCore import *
from Qt.QtWidgets import *

import tpDcc as tp

from tpDcc.libs.qt.core import qtutils, statusbar

import tpRigToolkit


class WindowStatusBar(statusbar.StatusWidget, object):
    def __init__(self, parent=None):
        super(WindowStatusBar, self).__init__(parent)

        self._info_url = None
        self._tool = None

        self.setFixedHeight(25)
        self._info_btn = QPushButton()
        self._info_btn.setIconSize(QSize(25, 25))
        self._info_btn.setSizePolicy(QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed))
        self._info_btn.setIcon(tp.ResourcesMgr().icon('info1'))
        self._info_btn.setStyleSheet('QWidget {background-color: rgba(255, 255, 255, 0); border:0px;}')

        self._bug_btn = QPushButton()
        self._bug_btn.setIconSize(QSize(25, 25))
        self._bug_btn.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        self._bug_btn.setIcon(tp.ResourcesMgr().icon('bug'))
        self._bug_btn.setStyleSheet('QWidget {background-color: rgba(255, 255, 255, 0); border:0px;}')

        self.main_layout.insertWidget(0, self._info_btn)
        self.main_layout.insertWidget(1, self._bug_btn)

        self._info_btn.clicked.connect(self._on_open_url)
        self._bug_btn.clicked.connect(self._on_send_bug)

    def set_info_url(self, url):
        """
        Sets the URL used to open tool info documentation web
        :param url: str
        """

        self._info_url = url

    def set_tool(self, tool):
        """

        :param tool:
        :return:
        """

        self._tool = tool

    def has_url(self):
        """
        Returns whether the URL documentation web is set or not
        :return: bool
        """

        if self._info_url:
            return True

        return False

    def has_tool(self):
        """
        Returns whether window has a tool attached or not
        :return: bool
        """

        if self._tool:
            return True

        return False

    def show_info(self):
        """
        Shows the info button of the status bar
        """

        self._info_btn.setVisible(True)

    def hide_info(self):
        """
        Hides the info button of the status bar
        """

        self._info_btn.setVisible(False)

    def show_bug(self):
        self._bug_btn.setVisible(True)

    def hide_bug(self):
        self._bug_btn.setVisible(False)

    def open_info_url(self):
        """
        Opens tool documentation URL in user web browser
        """

        if not self._project:
            return False

        if self._info_url:
            webbrowser.open_new_tab(self._info_url)

    def _on_send_bug(self):

        if not self._project:
            return False

        tpRigToolkit.ToolsMgr().run_tool(self._project, 'bugtracker', extra_args={'tool': self._tool})

    def _on_open_url(self):
        """
        Internal callback function that is called when the user presses the info icon button
        :return:
        """

        self.open_info_url()


class MainWindow(tp.Window, object):

    LOGO_NAME = None
    STATUS_BAR_WIDGET = WindowStatusBar

    def __init__(
            self,
            tool=None,
            name='MainWindow',
            title='MainWindow',
            size=(800, 535),
            fixed_size=False,
            parent=None,
            *args,
            **kwargs):

        self._tool = tool

        super(MainWindow, self).__init__(
            name=name,
            title=title,
            size=size,
            fixed_size=fixed_size,
            auto_run=True,
            frame_less=True,
            use_style=False,
            parent=parent,
            *args,
            **kwargs
        )

        if self.parent():
            for widget in self.parent().findChildren(QMainWindow):
                if widget is not self:
                    if widget.objectName() == self.objectName():
                        widget.close()

        screen_geo = QApplication.desktop().screenGeometry()
        screen_width = screen_geo.width()
        screen_height = screen_geo.height()
        new_width = self.width()
        new_height = self.height()
        if self.width() > screen_width:
            new_width = 500
        if self.height() > screen_height:
            new_height = 500
        self.resize(new_width, new_height)
        self.center()

    def ui(self):
        super(MainWindow, self).ui()

        window_icon = self._get_icon()
        self.setWindowIcon(window_icon)

        title_layout = QHBoxLayout()
        title_layout.setContentsMargins(0, 0, 0, 0)
        title_layout.setSpacing(0)
        title_layout.setAlignment(Qt.AlignTop | Qt.AlignHCenter)
        self.main_layout.insertLayout(0, title_layout)

        self._status_bar.set_tool(self._tool)
        if not self._status_bar.has_url():
            self._status_bar.hide_info()
        if not self._status_bar.has_tool():
            self._status_bar.hide_bug()

        if self._tool:
            self.main_layout.addWidget(self._tool)
        #
        # if self._project.is_dev():
        #     int_colors = self._project.dev_color0.split(',')
        #     dev_style = "background-color: rgb({}, {}, {})".format(
        #         int_colors[0], int_colors[1], int_colors[2], 255)
        #     self._dragger.setStyleSheet(dev_style)

    def setWindowTitle(self, title):
        # if self._project.is_dev():
        #     title = '{} - [{}]'.format(title, self._project.get_environment())

        super(MainWindow, self).setWindowTitle(title)

    def closeEvent(self, event):
        if self._tool:
            self._tool.close_plugin()
        super(MainWindow, self).closeEvent(event)

    def add_toolbar(self, name, area=Qt.TopToolBarArea):
        """
        Overrides base MainWindow add_toolbar function
        :param name: str
        :param area:
        :return:
        """

        new_toolbar = QToolBar(name)
        # The 0 widget is always the header view of the window
        self.main_layout.insertWidget(1, new_toolbar)
        return new_toolbar

    def set_info_url(self, url):
        """
        Sets the info URL of the current window
        :param url: str
        """

        if not url:
            return

        self._status_bar.set_info_url(url)

        if not self._status_bar.has_url():
            self._status_bar.hide_info()
        else:
            self._status_bar.show_info()

    def _get_icon(self):
        return tp.ResourcesMgr().icon('tprigtoolkit')


def dock_window(window_class, min_width=300):
    """
    Utility function to dock Maya window
    :param window_class: cls
    """

    if not tp.is_maya():
        return

    import maya.cmds as cmds
    import maya.OpenMayaUI as OpenMayaUI
    try:
        cmds.deleteUI(window_class.name)
    except Exception:
        pass

    main_control = cmds.workspaceControl(
        window_class.name, ttc=["AttributeEditor", -1], iw=min_width, mw=True, wp='preferred', label=window_class.title)

    control_widget = OpenMayaUI.MQtUtil.findControl(window_class.name)
    control_wrap = qtutils.wrapinstance(int(control_widget), QWidget)
    control_wrap.setAttribute(Qt.WA_DeleteOnClose)
    win = window_class(parent=control_wrap)

    cmds.evalDeferred(lambda *args: cmds.workspaceControl(main_control, e=True, rs=True))

    win.show()

    return win



# #! /usr/bin/env python
# # -*- coding: utf-8 -*-
#
# """
# Module that contains base class to handle tpRigToolkit packages
# """
#
# from __future__ import print_function, division, absolute_import
#
# import os
# import pkgutil
# import inspect
# import logging
#
# from tpDcc.libs.qt.core import preferences, tool
#
# from PyFlow.Core import NodeBase, FunctionLibrary, PinBase
#
# from tpRigToolkit.core import factory
# # from tpNodeGraph.core.base import functionlib, factory
# # from tpNodeGraph.core.data import node, port
#
# LOGGER = logging.getLogger('tpRigToolkit')
#
#
# class Package(object):
#     """
#     Base class that defines packages that can be used to extend tpNodeGraph functionality
#     """
#
#     def __init__(self, module_path):
#         super(Package, self).__init__()
#
#         self._package_path = module_path
#
#         self._node_classes = dict()
#         self._node_paths = dict()
#         self._port_classes = dict()
#         self._port_paths = dict()
#         self._function_libs = dict()
#         self._function_lib_paths = dict()
#         self._ui_nodes_factory_classes = dict()
#         self._ui_nodes_factory_paths = dict()
#         self._ui_ports_factory_classes = dict()
#         self._ui_ports_factory_paths = dict()
#         self._input_widgets_factory_classes = dict()
#         self._input_widgets_factory_paths = dict()
#         self._tool_classes = dict()
#         self._tool_paths = dict()
#         self._preference_widget_classes = dict()
#         self._preference_widget_paths = dict()
#
#         self.load()
#
#     @property
#     def package_path(self):
#         """
#         Returns path where module is located
#         :return: str
#         """
#
#         return self._package_path
#
#     def load(self):
#         """
#         Load info of the module
#         :return:
#         """
#
#         if not self._package_path or not os.path.isdir(self._package_path):
#             LOGGER.warning('Module Path is not valid: "{}"!'.format(self._package_path))
#             return
#         for importer, mod_name, is_pkg in pkgutil.walk_packages([self._package_path]):
#             module = importer.find_module(mod_name).load_module(mod_name)
#             for cname, obj in inspect.getmembers(module, inspect.isclass):
#                 obj_name = obj.__name__
#                 try:
#                     obj_path = inspect.getfile(obj)
#                 except Exception:
#                     continue
#                 if issubclass(obj, NodeBase):
#                     self._node_classes[obj_name] = obj
#                     self._node_paths[obj_name] = obj_path
#                 elif issubclass(obj, PinBase):
#                     self._port_classes[obj_name] = obj
#                     self._port_paths[obj_name] = obj_path
#                 elif issubclass(obj, FunctionLibrary.FunctionLibraryBase):
#                     self._function_libs[obj_name] = obj(module_name=self.__class__.__name__)
#                     self._function_lib_paths[obj_name] = obj_path
#                 elif issubclass(obj, factory.NodeFactory):
#                     self._ui_nodes_factory_classes[obj_name] = obj
#                     self._ui_nodes_factory_paths[obj_name] = obj_path
#                 elif issubclass(obj, factory.PinFactory):
#                     self._ui_ports_factory_classes[obj_name] = obj
#                     self._ui_ports_factory_paths[obj_name] = obj_path
#                 elif issubclass(obj, factory.InputWidgetsFactory):
#                     self._input_widgets_factory_classes[obj_name] = obj
#                     self._input_widgets_factory_paths[obj_name] = obj_path
#                 elif issubclass(obj, tool.BaseTool):
#                     self._tool_classes[obj_name] = obj
#                     self._tool_paths[obj_name] = obj_path
#                 elif issubclass(obj, preferences.CategoryWidgetBase):
#                     self._preference_widget_classes[obj_name] = obj
#                     self._preference_widget_paths[obj_name] = obj_path
#
#     @property
#     def node_classes(self):
#         """
#         Returns registered node classes
#         :return: list(dict(str, class))
#         """
#
#         return self._node_classes
#
#     @property
#     def node_paths(self):
#         """
#         Returns registered node paths
#         :return: list(dict(str, str))
#         """
#
#         return self._node_paths
#
#     @property
#     def port_classes(self):
#         """
#         Returns registered port classes
#         :return: list(dict(str, class))
#         """
#
#         return self._port_classes
#
#     @property
#     def port_paths(self):
#         """
#         Returns registered port paths
#         :return: list(dict(str, str))
#         """
#
#         return self._port_paths
#
#     @property
#     def function_libs(self):
#         """
#         Returns registered function libs instances
#         :return: list(dict(str, FunctionLibrary))
#         """
#
#         return self._function_libs
#
#     @property
#     def function_lib_paths(self):
#         """
#         Returns registered function lib paths
#         :return: list(dict(str, str))
#         """
#
#         return self._function_lib_paths
#
#     @property
#     def ui_nodes_factory_classes(self):
#         """
#         Returns registered UI nodes factory classes
#         :return: list(dict(str, str))
#         """
#
#         return self._ui_nodes_factory_classes
#
#     @property
#     def ui_nodes_factory_paths(self):
#         """
#         Returns registered UI nodes factory paths
#         :return: list(dict(str, str))
#         """
#
#         return self._ui_nodes_factory_paths
#
#     @property
#     def ui_ports_factory_classes(self):
#         """
#         Returns registered UI ports factory classes
#         :return: list(dict(str, str))
#         """
#
#         return self._ui_ports_factory_classes
#
#     @property
#     def ui_ports_factory_paths(self):
#         """
#         Returns registered UI ports factory paths
#         :return: list(dict(str, str))
#         """
#
#         return self._ui_ports_factory_paths
#
#     @property
#     def input_widgets_factory_classes(self):
#         """
#         Returns registered input widgets factory classes
#         :return: list(dict(str, str))
#         """
#
#         return self._input_widgets_factory_classes
#
#     @property
#     def tool_classes(self):
#         """
#         Returns registered tool classes
#         :return: list(dict(str, class))
#         """
#
#         return self._tool_classes
#
#     @property
#     def tool_paths(self):
#         """
#         Returns registered tool paths
#         :return: list(dict(str, str))
#         """
#
#         return self._tool_paths
#
#     @property
#     def preference_widget_classes(self):
#         """
#         Returns registered preference widget classes
#         :return: list(dict(str, class))
#         """
#
#         return self._preference_widget_classes
#
#     @property
#     def preference_widget_paths(self):
#         """
#         Returns registered preference widget paths
#         :return: list(dict(str, str)
#         """
#
#         return self._preference_widget_paths
#
#     def has_node_classes(self):
#         """
#         Returns whether current module has available nodes or not
#         :return: bool
#         """
#
#         return bool(self._node_classes)
#
#     def has_port_classes(self):
#         """
#         Returns whether current module has available ports or not
#         :return: bool
#         """
#
#         return bool(self._port_classes)
#
#     def has_ui_nodes_factory_classes(self):
#         """
#         Returns whether current module has available UI nodes factories or not
#         :return: bool
#         """
#
#         return bool(self._ui_nodes_factory_classes)
#
#     def has_function_lib_classes(self):
#         """
#         Returns whether current module has available function libraries or not
#         :return: bool
#         """
#
#         return bool(self._function_libs)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains implementation for node manager
"""

from __future__ import print_function, division, absolute_import

import os
import pkgutil
import logging
import traceback
from collections import defaultdict

from tpDcc.libs.python import python, decorators

from tpRigToolkit import packages
# from tpRigToolkit.core import package

LOGGER = logging.getLogger('tpRigToolkit')

"""
TODO
This is class is related to PyFlow packages implementation. We do not need altough before removing it we should
check if this can be useful for the RigNode implementation
"""


class PackagesManager(object):

    ENV_VAR_NAME = 'TPRIGTOOLKIT_PACKAGE_PATHS'
    ENV_VAR_PATHS_DELIMITER = ';'

    def __init__(self, package_paths=None):
        super(PackagesManager, self).__init__()

        self._registered_modules = dict()
        self._registered_module_paths = dict()
        self._registered_ui_nodes_factories = dict()
        self._registered_ui_ports_factories = dict()
        self._registered_input_widgets_factories = dict()
        self._registered_tools = defaultdict(list)
        self._registered_prefs = defaultdict(list)

        self._module_paths = python.force_list(package_paths) if package_paths else list()
        self._update_module_paths_from_environment()

    @property
    def registered_modules(self):
        """
        Returns dict with the registered modules
        :return: dict
        """

        return self._registered_modules

    @property
    def registered_modules_paths(self):
        """
        Returns dict containing registered modules paths
        :return: dict
        """

        return self._registered_module_paths

    @property
    def registered_ui_nodes_factories(self):
        """
        Returns dict containing registered UI nodes factories
        :return: dict
        """

        return self._registered_ui_nodes_factories

    @property
    def registered_ui_ports_factories(self):
        """
        Returns dict containing registered UI ports factories
        :return: dict
        """

        return self._registered_ui_ports_factories

    @property
    def registered_input_widgets_factories(self):
        """
        Returns dict containing registered input widgets
        :return: dict
        """

        return self._registered_input_widgets_factories

    @property
    def registered_tools(self):
        """
        Returns dict containing registered tools
        :return: dict
        """

        return self._registered_tools

    @property
    def registered_prefs(self):
        """
        Returns dict containing registered preference widgets
        :return: dict
        """

        return self._registered_prefs

    def register_module_path(self, module_path):
        """
        Registers given path into the manager so path is taking into account when loading tpNodeGrpah modules
        :param module_path: str
        """

        if not module_path or not os.path.isdir(module_path) or module_path in self._module_paths:
            return

        self._module_paths.append(module_path)
        self._load_modules(module_path)

    def register_ui_node_factory(self, module_name, factory):
        """
        Registers given factory and maps it to given module name
        :param module_name: str
        :param factory: Factory
        """

        if module_name not in self._registered_ui_nodes_factories:
            self._registered_ui_nodes_factories[module_name] = factory

    def register_ui_port_factory(self, module_name, factory):
        """
        Registers given factory and maps it to given module name
        :param module_name: str
        :param factory: Factory
        """

        if module_name not in self._registered_ui_ports_factories:
            self._registered_ui_ports_factories[module_name] = factory

    def register_input_widgets_factory(self, module_name, factory):
        """
        Registers given factory and maps it to given module name
        :param module_name: str
        :param factory: Factory
        """

        if module_name not in self._registered_input_widgets_factories:
            self._registered_input_widgets_factories[module_name] = factory

    def register_tool(self, module_name, tool_class):
        """
        Registers given tool class and maps it to given module name
        :param module_name: str
        :param tool_class: cls
        """

        registered_tool_names = [tool.NAME for tool in self._registered_tools[module_name]]
        if tool_class.NAME not in registered_tool_names:
            self._registered_tools[module_name].append(tool_class)
            tool_class.MODULE_NAME = module_name

    def register_preference_widget(self, module_name, preference_widget_class):
        """
        Registers given preference widget class and maps it to the given module name
        :param module_name: str
        :param preference_widget_class: cls
        """

        registered_prefs_categories = [pref.CATEGORY for pref in self._registered_prefs[module_name]]
        if preference_widget_class.CATEGORY not in registered_prefs_categories:
            self._registered_prefs[module_name].append(preference_widget_class)
            preference_widget_class.MODULE_NAME = module_name

    def initialize(self):
        import PyFlow
        self._load_modules()

    def _load_modules(self, module_path=None):
        """
        Function that loads all the modules located in the registered paths
        """

        if not module_path:
            module_paths = self._module_paths
        else:
            module_paths = [module_path]

        for importer, mod_name, is_pkg in pkgutil.iter_modules(module_paths):
            try:
                if is_pkg:
                    mod = importer.find_module(mod_name).load_module(mod_name)
                    mod_path = mod.__path__[0]
                    if hasattr(mod, 'MODULE_NAME'):
                        module_name = mod.MODULE_NAME
                    else:
                        module_name = mod_name
                    mod_class = type(module_name, (package.Package,), {})
                    new_module = mod_class(mod_path)
                    self._registered_modules[module_name] = new_module
                    self._registered_module_paths[module_name] = os.path.normpath(mod_path)
            except Exception as e:
                LOGGER.error('Error on Module: {} : \n{}'.format(mod_name, str(e)))
                LOGGER.error(traceback.format_exc())
                continue

        registered_internal_port_data_types = set()

        for name, mod in self._registered_modules.items():
            module_name = mod.__class__.__name__

            for n in mod.node_classes.values():
                n.MODULE_NAME = module_name

            for p in mod.port_classes.values():
                p.MODULE_NAME = module_name
                if p.is_value_port():
                    internal_type = p.internal_data_structure()
                    if internal_type in registered_internal_port_data_types:
                        raise Exception(
                            'Port with "{}" internal data type already been registered!'.format(internal_type))
                    registered_internal_port_data_types.add(internal_type)

            ui_nodes_factory = mod.ui_nodes_factory_classes
            if ui_nodes_factory:
                self.register_ui_node_factory(module_name, ui_nodes_factory)

            ui_ports_factory = mod.ui_ports_factory_classes
            if ui_ports_factory:
                self.register_ui_port_factory(name, ui_ports_factory)

            input_widgets_factory = mod.input_widgets_factory_classes
            if input_widgets_factory:
                self.register_input_widgets_factory(name, input_widgets_factory)

            for tool_class in mod.tool_classes.values():
                supported_softwares = tool_class.SUPPORTED_SOFTWARES
                if 'any' not in supported_softwares:
                    continue
                self.register_tool(name, tool_class)

            for pref_widget_class in mod.preference_widget_classes.values():
                self.register_preference_widget(name, pref_widget_class)

    def _update_module_paths_from_environment(self):
        """
        Internal function that updates registered module paths by taking into account current environment variables
        """

        def _recurse_module_paths(module_path):
            """
            Recursively loops through given path searching modules
            :param module_path: str
            :return: list(str)
            """

            paths_found = list()
            for sub_folder in os.listdir(module_path):
                sub_folder_path = os.path.join(module_path, sub_folder)
                if os.path.isdir(sub_folder_path):
                    if sub_folder_path.endswith('Package') or sub_folder_path.endswith('package'):
                        paths_found.append(sub_folder_path)

            return paths_found

        module_paths = packages.__path__

        if self.ENV_VAR_NAME in os.environ:
            extra_paths = os.environ.get(self.ENV_VAR_NAME).rstrip(self.ENV_VAR_PATHS_DELIMITER)
            for modules_root in extra_paths.split(self.ENV_VAR_PATHS_DELIMITER):
                if os.path.isdir(modules_root):
                    paths = _recurse_module_paths(modules_root)
                    extra_paths.extend(paths)

            module_paths.extend(extra_paths)

        for p in module_paths:
            self.register_module_path(p)


@decorators.Singleton
class PackagesManagerSingleton(PackagesManager, object):
    def __init__(self):
        PackagesManager.__init__(self)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains script editor tool implementation
"""

from __future__ import print_function, division, absolute_import

from Qt.QtCore import *
from Qt.QtWidgets import *

from tpDcc.tools.scripteditor.widgets import scripteditor

from tpRigToolkit.core import plugin


class ScriptsEditorPlugin(plugin.DockPlugin, object):

    NAME = 'Scripts Editor'
    TOOLTIP = 'Allow to edit scripts easily'
    DEFAULT_DOCK_AREA = Qt.RightDockWidgetArea
    IS_SINGLETON = True

    def __init__(self):
        super(ScriptsEditorPlugin, self).__init__()

        self._script_editor_widget = None
        self._content = QWidget()
        self._content_layout = QVBoxLayout()
        self._content_layout.setContentsMargins(0, 0, 0, 0)
        self._content_layout.setSpacing(0)
        self._content.setLayout(self._content_layout)
        self.setWidget(self._content)

    def show_plugin(self):
        super(ScriptsEditorPlugin, self).show_plugin()

        settings = self._app.settings()

        if not self._script_editor_widget:
            self._script_editor_widget = scripteditor.ScriptEditorWidget(settings=settings, load_session=False)
            # self._script_editor_widget.disable_save_script()
            self._script_editor_widget.disable_console()
            self._script_editor_widget.set_toolbar_visibility(False)
            self._script_editor_widget.set_menubar_visibility(False)
            self._script_editor_widget.close_all_tabs()
            self._script_editor_widget.scriptSaved.connect(self._on_script_saved)
            self._script_editor_widget.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
            self._content_layout.addWidget(self._script_editor_widget)
            self._script_editor_widget.lastTabClosed.connect(self.close)

    def load_script(self, script_file):
        if not self._script_editor_widget:
            return

        self._script_editor_widget.load_script(script_file)

    def _on_script_saved(self, file_path):
        pass



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains base class to create tpNodeGraph factories
"""

from __future__ import print_function, division, absolute_import


class Factory(object):
    """
    Base class that defines factory classes
    """

    def __init__(self):
        super(Factory, self).__init__()

    def create(self, raw_inst, *args, **kwargs):
        pass


class NodeFactory(Factory, object):
    """
    Base class that defines node factory class
    """

    def __init__(self):
        super(NodeFactory, self).__init__()


class PinFactory(Factory, object):
    """
    Base class that defines node factory class
    """

    def __init__(self):
        super(PinFactory, self).__init__()

    def create(self, raw_inst, owning_node):
        pass


class InputWidgetsFactory(object):
    """
    Base class that defines input widgets factory class
    """

    def __init__(self):
        super(InputWidgetsFactory, self).__init__()

    def create(self, data_type, data_setter, default_value=None, widget_variant=None, **kwargs):
        pass



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Initialization module for tpRigToolkit
"""

from __future__ import print_function, division, absolute_import

__author__ = "Tomas Poveda"
__license__ = "MIT"
__maintainer__ = "Tomas Poveda"
__email__ = "tpovedatd@gmail.com"

import os
import logging.config

import tpDcc
from tpRigToolkit import register

# =================================================================================

PACKAGE = 'tpRigToolkit'

# =================================================================================


def init(import_libs=True, dev=False):
    """
    Initializes tpRigToolkit module
    :param import_libs: bool, Whether to import deps libraries by default or not
    :param dev: bool, Whether tpRigToolkit is initialized in dev mode or not
    """

    if dev:
        register.cleanup()

    logger = create_logger(dev=dev)

    register.register_class('logger', logger)

    if import_libs:
        import tpDcc.loader as dcc_loader
        dcc_loader.init(dev=dev)

    # We do the importer and registering of the classes after ensuring tpDcc is loaded
    from tpRigToolkit.managers import data as data_manager, names as names_manager, packages as packages_manager
    from tpRigToolkit.managers import plugins as plugins_manager, scripts as scripts_manager
    from tpRigToolkit.widgets import window, dialog

    register.register_class('DataMgr', data_manager.DataManagerSingleton)
    register.register_class('NamesMgr', names_manager.RigToolkitNamesManagerSingleton)
    register.register_class('PackagesMgr', packages_manager.PackagesManagerSingleton)
    register.register_class('PluginsMgr', plugins_manager.PluginsManagerSingleton)
    register.register_class('ScriptsMgr', scripts_manager.ScriptsManagerSingleton)
    register.register_class('Window', window.MainWindow)
    register.register_class('Dialog', dialog.MainDialog)

    # skip_modules = ['{}.{}'.format(PACKAGE, name) for name in ['libs', 'tools']]
    # importer.init_importer(package=PACKAGE, skip_modules=skip_modules)

    init_managers(dev=dev)

    dcc_loader_module = tpDcc.get_dcc_loader_module('tpRigToolkit.dccs')
    if dcc_loader_module:
        dcc_loader_module.init(dev=dev)


def init_managers(dev=True):
    """
    Initializes all tpDcc managers
    """

    import tpDcc
    import tpRigToolkit
    from tpRigToolkit import config, toolsets

    # Register resources
    resources_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'resources')
    tpDcc.ResourcesMgr().register_resource(resources_path)

    # Register configuration files
    tpDcc.ConfigsMgr().register_package_configs('tpRigToolkit', os.path.dirname(config.__file__))

    core_config = tpDcc.ConfigsMgr().get_config(
        'tpRigToolkit-core',
        environment='development' if dev else 'production'
    )
    if not core_config:
        tpDcc.logger.warning(
            'tpRigToolkit-core configuration file not found! '
            'Make sure that you have tpRigToolkit-config package installed!')
        return None

    # Register resources
    resources_path = os.path.normpath(os.path.join(os.path.dirname(__file__), 'resources'))
    tpDcc.ResourcesMgr().register_resource(resources_path)

    # Register tools
    tools_to_load = core_config.get('tools', list())
    tpDcc.ToolsMgr().register_package_tools(pkg_name=PACKAGE, tools_to_register=tools_to_load, dev=dev)
    tpDcc.ToolsMgr().load_registered_tools(PACKAGE)

    # Create tpRigToolkit menu
    tpDcc.MenusMgr().create_menus(package_name=PACKAGE)

    # Register toolsets
    tpDcc.ToolsetsMgr().register_path(PACKAGE, os.path.dirname(os.path.abspath(toolsets.__file__)))
    tpDcc.ToolsetsMgr().load_registered_toolsets(package_name=PACKAGE, tools_to_load=tools_to_load)

    # Initialize data and scripts manager
    tpRigToolkit.DataMgr()
    tpRigToolkit.ScriptsMgr()


def create_logger(dev=False):
    """
    Returns logger of current module
    """

    logger_dir = os.path.normpath(os.path.join(os.path.expanduser('~'), 'tpRigToolkit', 'logs'))
    if not os.path.isdir(logger_dir):
        os.makedirs(logger_dir)

    logging_config = os.path.normpath(os.path.join(os.path.dirname(__file__), '__logging__.ini'))

    logging.config.fileConfig(logging_config, disable_existing_loggers=False)
    logger = logging.getLogger('tpRigToolkit')
    if dev:
        logger.setLevel(logging.DEBUG)
        for handler in logger.handlers:
            handler.setLevel(logging.DEBUG)

    return logger


if __name__ == '__main__':
    print('Hello World!')



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains project widgets implementation for tpRigToolkit
"""

from __future__ import print_function, division, absolute_import

from Qt.QtCore import *
from Qt.QtWidgets import *
from Qt.QtGui import *

import tpDcc as tp
from tpDcc.libs.qt.core import base
from tpDcc.libs.qt.widgets import dividers, buttons

import tpRigToolkit
from tpRigToolkit.widgets.options import rigoptionsviewer


class ProjectSettingsWidget(base.BaseWidget, object):
    exitSettings = Signal()

    def __init__(self, project=None, parent=None):
        self._project = None
        super(ProjectSettingsWidget, self).__init__(parent=parent)

        if project:
            self.set_project(project)

    def ui(self):
        super(ProjectSettingsWidget, self).ui()

        image_layout = QHBoxLayout()
        image_layout.setContentsMargins(2, 2, 2, 2)
        image_layout.setSpacing(2)
        self.main_layout.addLayout(image_layout)
        self._project_image = QLabel()
        self._project_image.setAlignment(Qt.AlignCenter)
        image_layout.addItem(QSpacerItem(30, 0, QSizePolicy.Expanding, QSizePolicy.Fixed))
        image_layout.addWidget(self._project_image)
        image_layout.addItem(QSpacerItem(30, 0, QSizePolicy.Expanding, QSizePolicy.Fixed))

        self.main_layout.addWidget(dividers.Divider('Nomenclature'))
        self._naming_widget = NamingWidget(project=self._project)
        self.main_layout.addWidget(self._naming_widget)

        self.main_layout.addWidget(dividers.Divider('Settings'))
        self._project_options_widget = rigoptionsviewer.RigOptionsViewer(option_object=self._project)
        self.main_layout.addWidget(self._project_options_widget)
        self.main_layout.addWidget(dividers.Divider())

        bottom_layout = QVBoxLayout()
        bottom_layout.setContentsMargins(2, 2, 2, 2)
        bottom_layout.setSpacing(2)
        bottom_layout.setAlignment(Qt.AlignBottom)
        self.main_layout.addLayout(bottom_layout)
        bottom_layout.addLayout(dividers.DividerLayout())

        buttons_layout = QHBoxLayout()
        buttons_layout.setContentsMargins(2, 2, 2, 2)
        buttons_layout.setSpacing(2)
        bottom_layout.addLayout(buttons_layout)

        ok_icon = tp.ResourcesMgr().icon('ok')
        back_icon = tp.ResourcesMgr().icon('back')
        self._ok_btn = buttons.StyleBaseButton(
            icon=ok_icon, icon_padding=2, parent=self, button_style=buttons.ButtonStyles.FlatStyle)
        self._ok_btn.setMinimumSize(QSize(35, 35))
        self._back_btn = buttons.StyleBaseButton(
            icon=back_icon, icon_padding=2, parent=self, button_style=buttons.ButtonStyles.FlatStyle)
        self._back_btn.setMaximumWidth(50)
        buttons_layout.addWidget(self._ok_btn)
        buttons_layout.addWidget(self._back_btn)

    def setup_signals(self):
        self._ok_btn.clicked.connect(self._on_save)
        self._back_btn.clicked.connect(self._on_exit)

    def get_project(self):
        """
        Returns current RigBuilder project used by this widget
        :return: Project
        """

        return self._project

    def set_project(self, project):
        """
        Sets current project used by this widget
        :param project: Project
        """

        self._project = project

        self._project_options_widget.set_option_object(self._project)
        if self._project:
            self._project_image.setPixmap(
                QPixmap(self._project.get_project_image()).scaled(QSize(150, 150), Qt.KeepAspectRatio))
        self._naming_widget.set_project(self._project)

    def update_options(self, do_reload=False):
        """
        Update options of the current project
        """

        if not self._project:
            return

        if do_reload:
            self._project.reload_options()

        self._project_options_widget.update_options()

    def _on_save(self):
        """
        Internal callback function that is called when the user exists settings widget
        """

        self.exitSettings.emit()

    def _on_exit(self):
        """
        Internal callback function that is called when the user exists settings widget
        """

        self.exitSettings.emit()


class NamingWidget(base.BaseWidget, object):
    def __init__(self, project=None, parent=None):

        self._project = project

        super(NamingWidget, self).__init__(parent=parent)

        self.update_rules()

    def get_main_layout(self):
        main_layout = QHBoxLayout()
        main_layout.setContentsMargins(0, 0, 0, 0)
        main_layout.setSpacing(2)

        return main_layout

    def ui(self):
        super(NamingWidget, self).ui()

        edit_icon = tp.ResourcesMgr().icon('edit')
        name_lbl = QLabel('Naming Rule: ')
        self._name_rules = QComboBox()
        self._name_rules.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        self._edit_btn = buttons.IconButton(icon=edit_icon, icon_padding=2, button_style=buttons.ButtonStyles.FlatStyle)
        self.main_layout.addWidget(name_lbl)
        self.main_layout.addWidget(self._name_rules)
        self.main_layout.addWidget(dividers.get_horizontal_separator_widget())
        self.main_layout.addWidget(self._edit_btn)

    def setup_signals(self):
        self._name_rules.currentIndexChanged.connect(self._on_update_rule)
        self._edit_btn.clicked.connect(self._on_open_naming_manager)

    def set_project(self, project):
        self._project = project
        self.update_rules()

    def update_rules(self):

        try:
            self._name_rules.blockSignals(True)

            self._name_rules.clear()
            if not self._project:
                return
            naming_lib = self._project.naming_lib
            if not naming_lib:
                return
            naming_lib.load_session()
            rules = naming_lib.rules
            for rule in rules:
                self._name_rules.addItem(rule.name, userData=rule)
            rule_name = self._set_rule()
            self._name_rules.setCurrentText(rule_name)
        except Exception as exc:
            tpRigToolkit.logger.warning('Error while updating rules: {}'.format(exc))
        finally:
            self._name_rules.blockSignals(False)

    def _set_rule(self, rule=None):
        if not self._project:
            return
        naming_lib = self._project.naming_lib
        if not naming_lib:
            return

        if rule:
            rule_name = rule.name
            if self._project.settings.has_setting('naming_rule'):
                current_rule = self._project.settings.get('naming_rule')
                if current_rule == rule_name:
                    return
                self._project.settings.set('naming_rule', rule_name)
        else:
            if not self._project.settings.has_setting('naming_rule'):
                self._project.settings.set('naming_rule', self._name_rules.currentText())

        if not self._project.settings.has_setting('naming_rule'):
            return
        rule_name = self._project.settings.get('naming_rule')
        if not naming_lib.has_rule(rule_name):
            return
        naming_lib.set_active_rule(rule_name)

        return rule_name

    def _on_update_rule(self, index):
        rule = self._name_rules.itemData(index)
        self._set_rule(rule=rule)

    def _on_open_naming_manager(self):
        naming_manager_tool = tp.ToolsMgr().launch_tool_by_id(
            'tpRigToolkit-tools-namemanager', do_reload=True, debug=False, project=self._project)
        attacher = naming_manager_tool.attacher
        attacher.closed.connect(self.update_rules)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains options to select rig bones/joints
"""

from __future__ import print_function, division, absolute_import

import os
import json

from Qt.QtCore import *
from Qt.QtWidgets import *

import tpDcc as tp
from tpDcc.libs.qt.core import base
from tpDcc.libs.qt.widgets import layouts, lineedit, buttons, dividers
from tpDcc.libs.qt.widgets.options import option, list, text

from tpRigToolkit.data import skeleton


class BoneOption(option.Option, object):
    def __init__(self, name, parent, main_widget):
        super(BoneOption, self).__init__(name=name, parent=parent, main_widget=main_widget)

    def get_option_type(self):
        return 'bone'

    def get_option_widget(self):
        return GetBoneWidget(name=self._name)

    def get_value(self):
        value = self._option_widget.get_text()
        if not value:
            value = ''

        return value

    def set_value(self, value):
        value = str(value)
        self._option_widget.set_text(value)

    def _setup_option_widget_value_change(self):
        self._option_widget.textChanged.connect(self._on_value_changed)


class BoneOptionList(list.ListOption, object):
    def __init__(self, name, parent, main_widget):
        super(BoneOptionList, self).__init__(name=name, parent=parent, main_widget=main_widget)

    def get_option_type(self):
        return 'boneList'


class GetBoneWidget(text.TextWidget, object):
    def __init__(self, name='', parent=None):
        super(GetBoneWidget, self).__init__(name=name, parent=parent)

    def get_text_widget(self):
        return BoneLineEdit()


class BoneLineEdit(lineedit.BaseLineEdit, object):
    def __init__(self, text='', parent=None):
        super(BoneLineEdit, self).__init__(text=text, parent=parent)

    @property
    def selected_node(self):
        return self.text()

    def dropEvent(self, event):
        if event.mimeData().hasUrls():
            data = event.mimeData().urls()
            file_info = QFileInfo(data[0].toLocalFile())
            file_name = file_info.absoluteFilePath()
            if file_name and os.path.isfile(file_name):
                file_extension = os.path.splitext(file_name)[-1]
                data_extension = skeleton.SkeletonFileData.get_data_extension()
                if not data_extension.startswith('.'):
                    data_extension = '.{}'.format(data_extension)
                if file_extension == data_extension:
                    self._show_bones_hierarchy(file_name)
                    event.accept()
        elif event.mimeData().hasText():
            self.setText(event.mimeData().text())

        event.accept()

    def _show_bones_hierarchy(self, file_path):
        dlg = QDialog(parent=tp.Dcc.get_main_window() or None)
        dlg.setWindowTitle('Select Skeleton Node')
        lyt = QVBoxLayout()
        lyt.setSpacing(0)
        lyt.setContentsMargins(0, 0, 0, 0)
        dlg.setLayout(lyt)
        bone_hierarchy_widget = BoneHierarchyWidget(file_path, parent=dlg)
        current_bone = self.text() or ''
        bone_hierarchy_widget.set_bone(current_bone)
        lyt.addWidget(bone_hierarchy_widget)
        dlg.exec_()
        selected_node = bone_hierarchy_widget.selected_node
        if not selected_node:
            return
        self.setText(selected_node)


class BoneHierarchyWidget(base.BaseWidget, object):
    def __init__(self, file_path, parent=None):
        self._file_path = file_path
        self._selected_node = None
        super(BoneHierarchyWidget, self).__init__(parent=parent)

        self._load_data()

    @property
    def selected_node(self):
        return self._selected_node

    def ui(self):
        super(BoneHierarchyWidget, self).ui()

        self._tree_hierarchy = QTreeWidget()
        self._node_line = QLineEdit()
        self._node_line.setReadOnly(True)
        self._ok_btn = buttons.BaseButton('Ok')
        self._cancel_btn = buttons.BaseButton('Cancel')
        buttons_layout = layouts.HorizontalLayout()
        buttons_layout.addStretch()
        buttons_layout.addWidget(self._ok_btn)
        buttons_layout.addWidget(self._cancel_btn)

        self.main_layout.addWidget(self._tree_hierarchy)
        self.main_layout.addWidget(self._node_line)
        self.main_layout.addWidget(dividers.Divider())
        self.main_layout.addLayout(buttons_layout)

    def setup_signals(self):
        self._tree_hierarchy.currentItemChanged.connect(self._on_item_selected)
        self._ok_btn.clicked.connect(self._on_ok)
        self._cancel_btn.clicked.connect(self._on_cancel)

    def set_bone(self, bone_name):
        if not bone_name:
            return

        find_items = self._tree_hierarchy.findItems(bone_name, Qt.MatchExactly | Qt.MatchRecursive, 0)
        if not find_items:
            return

        find_item = find_items[0]
        find_item.setSelected(True)
        self._tree_hierarchy.setCurrentItem(find_item)
        self._tree_hierarchy.scrollTo(self._tree_hierarchy.indexFromItem(find_item))

    def _load_data(self):
        self._tree_hierarchy.clear()
        if not self._file_path or not os.path.isfile(self._file_path):
            return

        with open(self._file_path, 'r') as fh:
            skeleton_data = json.load(fh)
        if not skeleton_data:
            return

        created_nodes = dict()
        for node_data in skeleton_data:
            node_index = node_data.get('index', 0)
            node_parent_index = node_data.get('parent_index', -1)
            node_name = node_data.get('name', 'new_node')
            new_node = QTreeWidgetItem()
            new_node.setText(0, node_name)
            created_nodes[node_index] = {'node': new_node, 'parent_index': node_parent_index}
        sorted(created_nodes.items(), key=lambda x: int(x[0]))

        for node_index, node_data in created_nodes.items():
            node_data = created_nodes.get(node_index, None)
            if not node_data:
                continue
            node_item = node_data.get('node')
            parent_index = node_data['parent_index']
            if parent_index <= -1:
                self._tree_hierarchy.addTopLevelItem(node_item)
                continue
            parent_node_data = created_nodes.get(parent_index, None)
            if not parent_node_data:
                continue
            parent_node_item = parent_node_data.get('node')
            parent_node_item.addChild(node_item)

        self._tree_hierarchy.expandAll()

    def _on_item_selected(self, current, previous):
        node_name = current.text(0)
        self._node_line.setText(node_name)
        self._selected_node = node_name

    def _on_ok(self):
        self.parent().close()

    def _on_cancel(self):
        self._selected_node = None
        self.parent().close()
